# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

import array
import errno
import os
from multiprocessing import Pipe
import re
import signal
import struct
import sys
import tempfile
import time

from nose.plugins.skip import SkipTest
from nose.tools import assert_equals, assert_true, assert_false, nottest

from cmf.monitor import MonitorDaemon, schema
from cmf.monitor.host import HostMonitor
from cmf.monitor.host.tcp_metrics import TcpMetrics
from cmf.monitor.host.filesystem_map import FilesystemMap, disk_usage_wrapper
from cmf.monitor.host.network_interfaces import NetworkInterfaces
from cmf.monitor.host_update import HostUpdate
from cmf.monitor.host.disk_devices import DiskDevice
from cmf.monitor.host.block_device_stats import BlockDeviceStats
from cmf.monitor.host.filesystem import Filesystem
from cmf.monitor.host.filesystems import Filesystems
from cmf.monitor.fs_update import FsUpdate

SAMPLE_BLOCK_DEVICE_STAT_1 = "66519761  2495668 1078049969 35954112 33291312 17183132 1041809864 121175252        0 12723812 157113448\n"
SAMPLE_BLOCK_DEVICE_STAT_2 = "66534428  2496485 1078307121 35961888 33295409 17187513 1041925392 121177676        0 12727416 157123632\n"
''' Sample calculations:
await_r = (35961888- 35954112) / (66534428-66519761) = 0.5301697688688893
await_w = (121177676- 121175252) / (33295409-33291312) = 0.5916524286062973

nr_ios = (66534428-66519761) + (33295409-33291312) = 18764
systm = (12727416 - 12723812) / nr_ios = 0.192069921126
'''
SAMPLE_AWAIT_READ_TIME = 0.5301697688688893
SAMPLE_AWAIT_WRITE_TIME = 0.5916524286062973
SAMPLE_SERVICE_TIME = 0.192069921126

def test_disk_await_calculation():
  stats1 = BlockDeviceStats([5, 0, 0, 10, 5, 0, 0, 10, 0, 0, 0])
  stats2 = BlockDeviceStats([10, 0, 0, 20, 10, 0, 0, 20, 0, 0, 0])
  assert_equals((2, 2), DiskDevice.calculate_await_times(stats2, stats1))

def test_disk_svctm_calculation():
  stats1 = BlockDeviceStats([5, 0, 0, 0, 5, 0, 0, 0, 0, 10, 0])
  stats2 = BlockDeviceStats([10, 0, 0, 0, 10, 0, 0, 0, 0, 30, 0])
  assert_equals(2, DiskDevice.calculate_service_time(stats2, stats1))

def test_disk_stat():
  test_file = tempfile.NamedTemporaryFile()
  try:
    test_file.write(SAMPLE_BLOCK_DEVICE_STAT_1)
    test_file.seek(0)
    device = DiskDevice('sda', stat_path=test_file.name)
    test_file.truncate()

    test_file.write(SAMPLE_BLOCK_DEVICE_STAT_2)
    test_file.seek(0)

    update = device.collect_update()
    metric_dict = { }
    for metric_obj in update.to_avro_dict()['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict.has_key(schema.DISK_AWAIT_READ_TIME))
    assert_true(metric_dict.has_key(schema.DISK_AWAIT_WRITE_TIME))
    assert_true(metric_dict.has_key(schema.DISK_SERVICE_TIME))
    assert_true(metric_dict.has_key(schema.DISK_READ_BYTES))
    assert_true(metric_dict.has_key(schema.DISK_WRITE_BYTES))
    assert_true(metric_dict.has_key(schema.DISK_READ_IOS))
    assert_true(metric_dict.has_key(schema.DISK_WRITE_IOS))
    assert_true(metric_dict[schema.DISK_SERVICE_TIME] - SAMPLE_SERVICE_TIME < 0.0001)
    assert_true(metric_dict[schema.DISK_AWAIT_READ_TIME] - SAMPLE_AWAIT_READ_TIME < 0.0001)
    assert_true(metric_dict[schema.DISK_AWAIT_WRITE_TIME] - SAMPLE_AWAIT_WRITE_TIME < 0.0001)
    assert_true(metric_dict[schema.DISK_READ_BYTES], 552093245952)
    assert_true(metric_dict[schema.DISK_WRITE_BYTES], 533465800704)
    assert_true(metric_dict[schema.DISK_READ_IOS], 66534428)
    assert_true(metric_dict[schema.DISK_WRITE_IOS], 33295409)
  finally:
    test_file.close()

SAMPLE_PROC_TCP = \
"""  sl  local_address rem_address   st tx_queue rx_queue tr tm->when retrnsmt   uid  timeout inode
   0: 00000000:C36E 00000000:0000 01 00000000:00000000 00:00000000 00000000   119        0 288886 1 0000000000000000 300 0 0 2 -1
   1: 0100007F:B4B1 00000000:0000 02 00000000:00000000 00:00000000 00000000   119        0 6560844 1 0000000000000000 300 0 0 2 -1
   2: 0100007F:ED33 00000000:0000 03 00000000:00000000 00:00000000 00000000  1000        0 15170726 1 0000000000000000 300 0 0 2 -1
   3: 0100007F:1F54 00000000:0000 04 00000000:00000000 00:00000000 00000000   118        0 7454264 1 0000000000000000 300 0 0 2 -1
   4: 0100007F:1F54 00000000:0000 05 00000000:00000000 00:00000000 00000000   118        0 7454264 1 0000000000000000 300 0 0 2 -1
1000: 0100007F:1F55 00000000:0000 06 00000000:00000000 00:00000000 00000000   119        0 286285 1 0000000000000000 300 0 0 2 -1"""

SAMPLE_PROC_TCP6 = \
"""  sl  local_address                         remote_address                        st tx_queue rx_queue tr tm->when retrnsmt   uid  timeout inode
   0: 00000000000000000000000000000000:1C0E 00000000000000000000000000000000:0000 07 00000000:00000000 00:00000000 00000000  1000        0 25410663 1 0000000000000000 300 0 0 2 -1
   1: 00000000000000000000000000000000:270E 00000000000000000000000000000000:0000 08 00000000:00000000 00:00000000 00000000   124        0 24776364 1 0000000000000000 300 0 0 2 -1
1000: 00000000000000000000000000000000:270E 00000000000000000000000000000000:0000 09 00000000:00000000 00:00000000 00000000   124        0 24776364 1 0000000000000000 300 0 0 2 -1"""

SAMPLE_PROC_NET_DEV = \
"""Inter-|   Receive                                                |  Transmit
 face |bytes    packets errs drop fifo frame compressed multicast|bytes    packets errs drop fifo colls carrier compressed
    lo: 29959835120 29778554    0    0    0     0          0        0    29959835120     29778554    0    0   0   0      0
  eth0: 20000000000 16000000    0    0    0     0    0   0    22200000000 16900000    0    0    0     0       0          0
"""

PROC_NET_DEV_PATTERN = "^\s*\w+:(\s+\d+){16}\s*$"

# Disabled until this works on Python 2.4
@nottest
def test_proc_net_dev_exists():
  if not sys.platform.startswith("linux"):
    raise SkipTest("Requires linux")

  proc_net_dev_pattern = re.compile(PROC_NET_DEV_PATTERN)
  try:
    dev_lines = open("/proc/net/dev", 'r').readlines()[2:]
    for dev_line in filter(None, [line.strip() for line in dev_lines]):
      assert_true(proc_net_dev_pattern.match(dev_line))
  except IOError:
    assert_true(False)

def test_nic_mode():
  test_file = tempfile.NamedTemporaryFile()
  try:
    test_file.write(SAMPLE_PROC_NET_DEV)
    test_file.seek(0)

    network_interfaces = NetworkInterfaces(proc_net_dev = test_file.name)
    network_interfaces.ioctl_ethtool = _ioctl_ethtool_mock_slow
    network_interfaces.collect()

    nics = network_interfaces.get_nic_updates()
    metric_dict = { }
    for metric_obj in nics[0].to_avro_dict()['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
  
    assert_true(metric_dict.has_key(schema.NETWORK_INTERFACE_SPEED))
    assert_true(metric_dict.has_key(schema.NETWORK_INTERFACE_DUPLEX_MODE))
  
    assert_equals(metric_dict[schema.NETWORK_INTERFACE_SPEED], 5)
    assert_equals(metric_dict[schema.NETWORK_INTERFACE_DUPLEX_MODE], 1)
  finally:
    test_file.close()

def test_ethtool_errors():
  for ethtool_fn in [_ioctl_ethtool_unsupported,
                     _ioctl_ethtool_inval,
                     _ioctl_ethtool_unknown_ioerror,
                     _ioctl_ethtool_unknown_error]:
    test_file = tempfile.NamedTemporaryFile()
    try:
      test_file.write(SAMPLE_PROC_NET_DEV)
      test_file.seek(0)

      network_interfaces = NetworkInterfaces(proc_net_dev = test_file.name)
      network_interfaces.ioctl_ethtool = ethtool_fn
      network_interfaces.collect()

      nics = network_interfaces.get_nic_updates()
      metric_dict = { }
      for metric_obj in nics[0].to_avro_dict()['metrics']:
        metric_dict[metric_obj['id']] = metric_obj['value']

      assert_false(metric_dict.has_key(schema.NETWORK_INTERFACE_SPEED))
      assert_false(metric_dict.has_key(schema.NETWORK_INTERFACE_DUPLEX_MODE))
      assert_equals(20000000000, metric_dict[schema.NETWORK_INTERFACE_BYTES_RECEIVE])
      assert_equals(16000000, metric_dict[schema.NETWORK_INTERFACE_PACKETS_RECEIVE])
      assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_ERRS_RECEIVE])
      assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_DROP_RECEIVE])
      assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_FRAME_RECEIVE])
      assert_equals(22200000000, metric_dict[schema.NETWORK_INTERFACE_BYTES_TRANSMIT])
      assert_equals(16900000, metric_dict[schema.NETWORK_INTERFACE_PACKETS_TRANSMIT])
      assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_ERRS_TRANSMIT])
      assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_DROP_TRANSMIT])
      assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_COLLS_TRANSMIT])
    finally:
      test_file.close()

def test_network_interfaces():
  # First test the constructor throws if the file doesn't exist
  try:
    NetworkInterfaces(proc_net_dev="/file/doesnt/exist/orimsurprisedifitdoes12345")
  except IOError, ioex:
    assert_true(ioex.errno == errno.ENOENT)

  test_file = tempfile.NamedTemporaryFile()
  try:
    test_file.write(SAMPLE_PROC_NET_DEV)
    test_file.seek(0)
  
    network_interfaces = NetworkInterfaces(proc_net_dev = test_file.name)
    network_interfaces.ioctl_ethtool = _ioctl_ethtool_mock
    network_interfaces.collect()

    nics = network_interfaces.get_nic_updates()
    # The number of nics returned is 1 if the loopback device was excluded via the regex
    assert_equals(len(nics), 1)
    avro_nic = nics[0].to_avro_dict()
    assert_equals(avro_nic['iface'], "eth0")
  
    metric_dict = { }
    for metric_obj in avro_nic['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
  
    assert_equals(20000000000, metric_dict[schema.NETWORK_INTERFACE_BYTES_RECEIVE])
    assert_equals(16000000, metric_dict[schema.NETWORK_INTERFACE_PACKETS_RECEIVE])
    assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_ERRS_RECEIVE])
    assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_DROP_RECEIVE])
    assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_FRAME_RECEIVE])
    assert_equals(22200000000, metric_dict[schema.NETWORK_INTERFACE_BYTES_TRANSMIT])
    assert_equals(16900000, metric_dict[schema.NETWORK_INTERFACE_PACKETS_TRANSMIT])
    assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_ERRS_TRANSMIT])
    assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_DROP_TRANSMIT])
    assert_equals(0, metric_dict[schema.NETWORK_INTERFACE_COLLS_TRANSMIT])
    assert_equals(1234, metric_dict[schema.NETWORK_INTERFACE_SPEED])
    assert_equals(1, metric_dict[schema.NETWORK_INTERFACE_DUPLEX_MODE])
  finally:
    test_file.close()

def test_tcp_connection_updates():
  test_tcp = tempfile.NamedTemporaryFile()
  try:
    test_tcp6 = tempfile.NamedTemporaryFile()
    try:
      test_tcp.write(SAMPLE_PROC_TCP)
      test_tcp.seek(0)
      test_tcp6.write(SAMPLE_PROC_TCP6)
      test_tcp6.seek(0)
      update = HostUpdate("/var/log", "/var/run")
      tcp_metrics = TcpMetrics(test_tcp.name, test_tcp6.name)
      tcp_metrics.add_tcp_metrics(update)

      avro_dict = update.to_avro_dict()
      metric_dict = { }
      for metric_obj in avro_dict['metrics']:
        metric_dict[metric_obj['id']] = metric_obj['value']

      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_ESTABLISHED], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_SYN_SENT], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_SYN_RECV], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_FIN_WAIT1], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_FIN_WAIT2], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_TIME_WAIT], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_CLOSE], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_CLOSE_WAIT], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_LAST_ACK], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_LISTEN], 0)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_CLOSING], 0)

    finally:
      test_tcp6.close()
  finally:
    test_tcp.close()

def test_filesystems_filter():
  class MockFilesystemMap:
    def get_map(self):
      return {'/data1': MockFilesystem('/data1'),
              '/data2': MockFilesystem('/data2')}

  class MockFilesystem:
    def __init__(self, mount):
      self._mount = mount
    def collect(self):
      pass
    def get_fs_update(self):
      return self._mount

  filesystems = Filesystems(MockFilesystemMap())
  filesystems.set_fs_exclude_regex('^\/data1$')
  filesystems.collect()
  updates = filesystems.get_fs_updates()
  assert_equals(1, len(updates))
  assert_equals('/data2', updates[0])

def test_filesystem_mount_walk():
  fs_map = FilesystemMap()

  # Note that the path doesn't need to exist
  my_path = "/tmp/foo/bar"
  my_fs = fs_map.get_filesystem_by_path(my_path)

  assert_true(my_fs is not None)
  assert_equals(my_fs, fs_map.get_filesystem_by_path(os.path.dirname(my_path)))

def test_host_update():
  daemon = MonitorDaemon( "host", "host", agent=None)
  reg = daemon.get_registry_for_testing()
  monitor = HostMonitor(daemon, reg)

  monitor.collect()
  avro = monitor.get_host_update().to_avro_dict()

  # Test fs update
  avro_fs_list = avro['fsUpdates']
  assert_true(len(avro_fs_list) >= 1, "You should have a local filesystem")
  fs1 = avro_fs_list[0]

  assert_true(os.path.isdir(fs1['mountpoint']))
  for metric in fs1['metrics']:
    if metric['id'] == schema.FS_CAPACITY:
      assert_true(metric['value'] > 0)
      break
  else:
    assert_true(False, "Metric list does not have the required metric")

  # Test that metrics are populated.
  by_id = dict( (k["id"], k["value"]) for k in avro["metrics"] )
  for m in [schema.LOAD_1, schema.LOAD_5, schema.LOAD_15, schema.CPU_PERCENT,
      schema.PHYSICAL_MEMORY_USED, schema.PHYSICAL_MEMORY_CACHED,
      schema.PHYSICAL_MEMORY_BUFFERS, schema.PHYSICAL_MEMORY_TOTAL,
      schema.TOTAL_CPU_USER, schema.TOTAL_CPU_SYSTEM, schema.TOTAL_CPU_IDLE,
      schema.TOTAL_CPU_NICE, schema.TOTAL_CPU_IOWAIT, schema.TOTAL_CPU_IRQ,
      schema.TOTAL_CPU_SOFT_IRQ]:
    assert_true(m in by_id, str(m) + " is missing")

def _ioctl_ethtool_mock(sockfd, iface, ecmd):
  # Return speed=1234 and duplex=1, the speed_hi bits are left as 0
  return array.array('B', struct.pack('12xHB29x', 1234, 1))

def _ioctl_ethtool_unsupported(sockfd, iface, ecmd):
  raise IOError(errno.EOPNOTSUPP, "Test- NIC unsupported!!")

def _ioctl_ethtool_inval(sockfd, iface, ecmd):
  raise IOError(errno.EINVAL, "Test- ethtool invalid arg, handle as unsupported!!")

def _ioctl_ethtool_unknown_ioerror(sockfd, iface, ecmd):
  raise IOError(errno.EIO, "Test- unknown IOError, but don't crash!!")

def _ioctl_ethtool_unknown_error(sockfd, iface, ecmd):
  raise Exception("Test- unknown exception, but don't crash!!")

def _ioctl_ethtool_mock_slow(sockfd, iface, ecmd):
  # Return speed=5 and duplex=1, the speed_hi bits are left as 0
  # Given the speed and duplex mode, the "slow mode" metric should be 1
  return array.array('B', struct.pack('12xHB29x', 5, 1))

def test_filesystem_map_disk_usage_wrapper():
  parent_pipe, child_pipe = Pipe(False)
  try:
    disk_usage_wrapper(child_pipe, ["/", "/tmp"])
    usage = parent_pipe.recv()
    assert_true(usage is not None)
    assert_true('/' in usage)
    assert_true('/tmp' in usage)
    assert_true('total' in usage['/'])
    assert_true('used' in usage['/'])
    assert_true('free' in usage['/'])
  finally:
    parent_pipe.close()
    child_pipe.close()

def test_filesystem_map_get_info_for_scm_heartbeat():
  class MockFilesystemMap(FilesystemMap):
    def __init__(self):
      FilesystemMap.__init__(self)
      self._fs_map = {}
      self._fs_map['/a'] = \
          Filesystem(self, '/a', '/a-partition', 'rw', False, 'blah')
      self._fs_map['/b'] = \
          Filesystem(self, '/b', '/b-partition', 'rw', True, 'blah')
      self._fs_map['/c'] = \
          Filesystem(self, '/c', '/c-partition', 'ro', False, 'blah')

    def disk_usage(self, mountpoint):
      if mountpoint == '/a':
        return dict(total=10, used=6, free=4)
      if mountpoint == '/b':
        return dict(total=20, used=12, free=8)
      if mountpoint == '/c':
        return dict(total=30, used=18, free=12)
      raise Exception("Unexpected mountpoint %s!" % mountpoint)

  info = MockFilesystemMap().get_info_for_scm_heartbeat()
  filesystem_details = info['filesystem_info']
  assert_equals(2, len(filesystem_details))
  seen = {}
  for fs in filesystem_details:
    if fs['name'] == '/b-partition':
      assert_equals('/b', fs['mount_point'])
      assert_equals(20, fs['total_space_bytes'])
      assert_equals(True, fs['nodev'])
    if fs['name'] == '/a-partition':
      assert_equals('/a', fs['mount_point'])
      assert_equals(10, fs['total_space_bytes'])
      assert_equals(False, fs['nodev'])
    seen[fs['name']] = True
  assert_equals(2, len(seen))

  free_space_info = info['mountpoint_available_space_bytes']
  assert_equals(2, len(free_space_info))
  assert_equals(4, free_space_info['/a'])
  assert_equals(8, free_space_info['/b'])
  
def test_filesystem_map():
  info = FilesystemMap().get_info_for_scm_heartbeat()
  assert_true(info is not None)
  filesystem_details = info['filesystem_info']
  assert_true(len(filesystem_details) > 0)

def test_disk_usage_for_nodev_filesystem_bad_function_call():
  def _sleep_function(pipe, paths):
    time.sleep(60.0)

  class MockFilesystemMap(FilesystemMap):
    def __init__(self):
      FilesystemMap.__init__(self, tolerance=10000000.0)
      self._fs_map = {}
      self._fs_map['/tmp'] = Filesystem(self, '/tmp', '/', 'rw', True, 'blah')

    def get_nodev_disk_usage_function(self):
      return _sleep_function

  fs_map = MockFilesystemMap()
  usage = fs_map.disk_usage('/tmp')
  assert_true(usage is None)
  assert_true(fs_map.nodev_refresh_process is None)

def test_disk_usage_for_nodev_filesystem_really_bad_function_call():
  def _sleep_function_with_signal_handler(pipe, paths):
    def _do_nothing_handler(signum, frame):
      pass
    original_handler = signal.signal(signal.SIGTERM, _do_nothing_handler)
    time.sleep(60.0)
    # The first signal will wake us up. Sleep again.
    time.sleep(60.0)
    
  class MockFilesystemMap(FilesystemMap):
    def __init__(self):
      FilesystemMap.__init__(self, tolerance=10000000.0)
      self._fs_map = {}
      self._fs_map['/tmp'] = Filesystem(self, '/tmp', '/', 'rw', True, 'blah')
      self.get_nodev_disk_usage_function_calls = 0

    def get_nodev_disk_usage_function(self):
      self.get_nodev_disk_usage_function_calls += 1
      return _sleep_function_with_signal_handler

  fs_map = MockFilesystemMap()
  usage = fs_map.disk_usage('/tmp')
  assert_true(usage is None)
  assert_true(fs_map.nodev_refresh_process is not None)
  assert_equals(1, fs_map.get_nodev_disk_usage_function_calls)
  usage = fs_map.disk_usage('/tmp')
  assert_true(usage is None)
  assert_true(fs_map.nodev_refresh_process is not None)
  assert_equals(1, fs_map.get_nodev_disk_usage_function_calls)

def test_disk_usage_for_nodev_filesystem_cache():
  class MockFilesystemMap(FilesystemMap):
    def __init__(self):
      FilesystemMap.__init__(self, tolerance=10000000.0)
      self._fs_map = {}
      self._fs_map['/tmp'] = Filesystem(self, '/tmp', '/', 'rw', True, 'blah')
      self.get_nodev_disk_usage_function_calls = 0

    def get_nodev_disk_usage_function(self):
      self.get_nodev_disk_usage_function_calls += 1      
      return FilesystemMap.get_nodev_disk_usage_function(self)

  fs_map = MockFilesystemMap()
  assert_equals(0, fs_map.get_nodev_disk_usage_function_calls)
  usage = fs_map.disk_usage('/tmp')
  assert_equals(1, fs_map.get_nodev_disk_usage_function_calls)
  assert_true(usage is not None)
  # Another call should hit the cache.
  usage = fs_map.disk_usage('/tmp')
  assert_equals(1, fs_map.get_nodev_disk_usage_function_calls)
  assert_true(usage is not None)
  # If we drop the tolerance, then we should not hit the cache
  fs_map.tolerance = -1.0
  usage = fs_map.disk_usage('/tmp')
  assert_equals(2, fs_map.get_nodev_disk_usage_function_calls)
  assert_true(usage is not None)
